# ***************************************************************
# Copyright (c) 2021 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
from .test_core import expect_error
import numpy as np
from jittor import init, Module
import numpy as np


@unittest.skipIf(not jt.compiler.has_acl, "No ACL found")
class TestACL(unittest.TestCase):

    @jt.flag_scope(use_acl=1)
    def test_array(self):
        print("use_acl", jt.flags.use_acl)
        a = jt.array([1,2,3])
        np.testing.assert_allclose(a.numpy(), [1,2,3])

    @jt.flag_scope(use_acl=1)
    def test_add(self):
        a = jt.array([1,2,3])
        b = a+a
        np.testing.assert_allclose(b.numpy(), [2,4,6])

    @jt.flag_scope(use_acl=1)
    def test_add_float(self):
        a = jt.array([1.0,2.0,3.0])
        b = a+a
        np.testing.assert_allclose(b.numpy(), [2,4,6])

    @jt.flag_scope(use_acl=1)
    def test_array_cast(self):
        # this test cannot pass because cast error
        x = np.random.rand(10)
        y = jt.float32(x)
        np.testing.assert_allclose(x, y.numpy())

    def test_meminfo(self):
        jt.display_memory_info()


def matmul(a, b):
    (n, m), k = a.shape, b.shape[-1]
    a = a.broadcast([n,m,k], dims=[2])
    b = b.broadcast([n,m,k], dims=[0])
    return (a*b).sum(dim=1)

class Linear(Module):
    def __init__(self, in_features, out_features, bias=True):
        self.w = (jt.random((in_features, out_features))-0.5) / in_features**0.5
        self.b = jt.random((out_features,))-0.5 if bias else None
    def execute(self, x):
        x = matmul(x, self.w)
        if self.b is not None: 
            return x+self.b
        return x

def relu(x):
    return jt.maximum(x, 0.0)
Relu = jt.make_module(relu)

class Model(Module):
    def __init__(self, input_size):
        self.linear1 = Linear(input_size, 10)
        self.relu1 = Relu()
        self.linear2 = Linear(10, 1)
    def execute(self, x):
        x = self.linear1(x)
        x = self.relu1(x)
        return self.linear2(x)

@unittest.skipIf(not jt.compiler.has_acl, "No ACL found")
class TestExample(unittest.TestCase):
    @jt.flag_scope(use_acl=1)
    def test1(self):
        np.random.seed(0)
        jt.set_seed(3)
        n = 1000
        batch_size = 50
        lr = 0.05

        def get_data(n):
            for i in range(n):
                x = np.random.rand(batch_size, 1).astype("float32")
                y = x*x
                yield jt.float32(x), jt.float32(y)
        
        model = Model(input_size=1)
        ps = model.parameters()

        for i,(x,y) in enumerate(get_data(n)):
            jt.sync_all(True)
            pred_y = model(x).name("pred_y")
            loss = ((pred_y - y).sqr()).name("loss")
            loss_mean = loss.mean()
            
            gs = jt.grad(loss_mean, ps)
            for p, g in zip(ps, gs):
                p -= g * lr

            if i>2:
                assert prev == jt.liveness_info(), f"memory leak {prev} {jt.liveness_info()}"
            prev = jt.liveness_info()
            print(f"step {i}, loss = {loss_mean.data.sum()} {jt.liveness_info()}")

        possible_results = [
            0.0009948202641680837,
            0.001381353591568768,
            0.00110957445576787,
        ]
        loss_mean = loss_mean.data
        assert any(abs(loss_mean - r) < 1e-6 for r in possible_results)

        jt.clean()

if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2021 Jittor. All Rights Reserved. 
# Maintainers: 
#     Wenyang Zhou <576825820@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import os, sys
import jittor as jt
import jittor_utils as jit_utils

class TestLock(unittest.TestCase):
    def test(self):
        if os.environ.get('lock_full_test', '0') == '1':
            cache_path = os.path.join(jit_utils.home(), ".cache", "jittor", "lock")
            assert os.system(f"rm -rf {cache_path}") == 0
            cmd = f"cache_name=lock {sys.executable} -m jittor.test.test_example"
        else:
            cmd = f"{sys.executable} -m jittor.test.test_example"
        print("run cmd twice", cmd)
        assert os.system(f"{cmd} & {cmd} & wait %1 && wait %2") == 0


if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2020 Jittor. Authors: 
#     Guowei Yang <471184555@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np

skip_this_test = False
try:
    jt.dirty_fix_pytorch_runtime_error()
    import torch
except:
    skip_this_test = True


@unittest.skipIf(skip_this_test, "No Torch Found")
class TestSearchsorted(unittest.TestCase):
    def test_searchsorted_cpu(self):
        for i in range(1,3):
            s = np.sort(np.random.rand(*((10,)*i)),-1)
            v = np.random.rand(*((10,)*i))
            s_jt = jt.array(s)
            v_jt = jt.array(v)
            s_tc = torch.from_numpy(s)
            v_tc = torch.from_numpy(v)

            y_tc = torch.searchsorted(s_tc, v_tc, right=True)
            y_jt = jt.searchsorted(s_jt, v_jt, right=True)
            assert np.allclose(y_jt.numpy(), y_tc.data)
            y_jt = jt.searchsorted(s_jt, v_jt, right=False)
            y_tc = torch.searchsorted(s_tc, v_tc, right=False)
            assert np.allclose(y_jt.numpy(), y_tc.data)

    @unittest.skipIf(not jt.compiler.has_cuda, "No CUDA found")
    @jt.flag_scope(use_cuda=1)
    def test_searchsorted_gpu(self):
        self.test_searchsorted_cpu()

if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2021 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
from .test_core import expect_error

def gen_data(shape):
    num = np.multiply.reduce(shape)
    a = np.arange(0, num)
    return a.reshape(shape).astype("int32")

    
class TestReduceOp(unittest.TestCase):
    def setUp(self):
        self.keepdims = False
        
    def test1(self):
        def check(a, op, dims):
            if ("logical" in op) and jt.flags.use_cuda:
                # TODO: atomic bool operation for cuda not 
                # supported yet
                return
            np_dims = jt_dims = dims
            if dims == ():
                np_dims = tuple(range(len(a.shape)))
            x = eval(f"np.{op}.reduce(a, {np_dims}, keepdims={self.keepdims})")
            y = eval(f"jt.reduce_{op}(a, {jt_dims}, keepdims={self.keepdims}).data")
            if len(x.shape) == 0: x = np.array([x]).astype(a.dtype)
            x = x.astype(a.dtype)
            y = y.astype(a.dtype)
            assert x.dtype == y.dtype and x.shape == y.shape and (x==y).all(), \
                f"\n{a.shape}\n{op}\n{dims}\n{x}\n{y}\n{x.dtype}\n{y.dtype}\n{a.dtype}"
                
        ia = [gen_data([2,3,4,5]), gen_data([5,3])]
        idims = [(), (0,), (1,), (2,), (3,), (0, 2), (1,3), (1,2,3), 2, 3]
        
        iop = [ op[7:] for op in dir(jt) if op.startswith("reduce_")]
        assert len(iop) >= 10, iop
        for a in ia:
            check(a, iop[0], idims[0])
        for op in iop:
            check(ia[0], op, idims[0])
        for dims in idims:
            check(ia[0], iop[0], dims)
        expect_error(lambda: jt.reduce_add([1,2,3], 2))

    def test_bool_reduce(self):
        x = (jt.bool([1,0,1]) | jt.bool([0,1,0])).all().item()
        assert x
        x = (jt.bool([1,0,1]) & jt.bool([0,1,0])).any().item()
        assert not x

    def test_bool_reduce2(self):
        def gen_data(shape):
            num = np.multiply.reduce(shape)
            a = np.random.randint(2, size=[num]).astype(bool)
            return a.reshape(shape).astype("int32")

        def check(a, op, dims):
            if ("logical" in op) and jt.flags.use_cuda:
                # TODO: atomic bool operation for cuda not 
                # supported yet
                return
            np_dims = jt_dims = dims
            if dims == ():
                np_dims = tuple(range(len(a.shape)))
            x = eval(f"np.{op}.reduce(a, {np_dims}, keepdims={self.keepdims})")
            y = eval(f"jt.reduce_{op}(a, {jt_dims}, keepdims={self.keepdims}).data")
            if len(x.shape) == 0: x = np.array([x]).astype(a.dtype)
            x = x.astype(a.dtype)
            y = y.astype(a.dtype)
            assert x.dtype == y.dtype and x.shape == y.shape and (x==y).all(), \
                f"\n{a.shape}\n{op}\n{dims}\n{x}\n{y}\n{x.dtype}\n{y.dtype}\n{a.dtype}"
                
        ia = [gen_data([2,3,4,5]), gen_data([5,3])]
        idims = [(), (0,), (1,), (2,), (3,), (0, 2), (1,3), (1,2,3), 2, 3]
        
        iop = [ op[7:] for op in dir(jt) if op.startswith("reduce_")]
        assert len(iop) >= 10, iop
        for a in ia:
            check(a, iop[0], idims[0])
        for op in iop:
            check(ia[0], op, idims[1])
        for dims in idims:
            check(ia[0], iop[0], dims)
        expect_error(lambda: jt.reduce_add([1,2,3], 2))


class TestReduceOp2(TestReduceOp):
    def setUp(self):
        self.keepdims = True


@unittest.skipIf(not jt.compiler.has_cuda, "No CUDA found")
class TestReduceOpCuda(TestReduceOp):
    def setUp(self):
        jt.flags.use_cuda = 2
        self.keepdims = False
    def tearDown(self):
        jt.flags.use_cuda = 0

@unittest.skipIf(not jt.compiler.has_cuda, "No CUDA found")
class TestReduceOpCuda2(TestReduceOp):
    def setUp(self):
        jt.flags.use_cuda = 2
        self.keepdims = True
    def tearDown(self):
        jt.flags.use_cuda = 0


class TestReduceOpMisc(unittest.TestCase):
    def test_negtive_dim(self):
        a = jt.array([[1,2],[3,4]])
        assert (a.sum(-1).data == [3,7]).all()
        assert (a.sum(-2).data == [4,6]).all()

if __name__ == "__main__":
    unittest.main()



# how to run:
# docker run -v "${HOME}/Documents/jittor-blog":/srv/jittor-blog -v /home/jittor/Documents/site:/mnt/jittor-blog -e LC_ALL=C.UTF-8 --rm jittor-blog-compiler bash -c "jekyll build --baseurl=JITTOR_BASEURL -d /mnt/jittor-blog/ && chmod -R 777 /mnt/jittor-blog"
# python /home/jittor/Documents/jittor-blog/local_doc_builder.py

import os

os.chdir("/home/jittor/Documents/site")

def check(dirname, fname):
    with open(os.path.join(dirname, fname), 'r') as f:
        src = f.read()
    ac = "JITTOR_BASEURL"
    rep = (
        ("href=\"//", "href=\"http://"), 
        ("src=\"//", "src=\"http://"),
        ('https://cg.cs.tsinghua.edu.cn/jittor', ac)
    )
    found = False
    for a,b in rep:
        if a in src:
            src = src.replace(a, b)
            found = True
    if ac not in src and not found: return
    n = len(dirname.split(os.path.sep))-1
    s = '.' + '/..' * n
    new_src = ""
    i = -1
    print("="*20)
    print(dirname, fname)
    while True:
        i += 1
        if i >= len(src):
            break
        if src[i] != 'J':
            new_src += src[i]
            continue
        if src[i:i+len(ac)] != ac:
            new_src += src[i]
            continue
        j = i
        while j<len(src) and src[j] != ' ' and src[j] != '"' and src[j] != "'":
            j += 1
        x = src[i:j]
        y = x.replace(ac, s)
        if '#' in y:
            y, l = y.split('#')
            l = '#'+l
        else:
            l = ""
        # replace xx/xx/ --> xx/xx/index.html
        if y.endswith('/'):
            y += 'index.html'
        else:
            z = y.split('/')[-1]
            # replace xx/xx --> xx/xx/index.html
            if '.' not in z:
                y += '/index.html'
        y += l
        print("found", x, '-->', y)
        new_src += y
        i = j-1
    with open(os.path.join(dirname, fname), 'w') as f:
        f.write(new_src)

for r, _, f in os.walk('.'):
    for fname in f:
        ext = fname.split('.')[-1]
        if ext not in ['html', 'css', 'js']:
            continue
        # print(r, fname)
        check(r, fname)




# ***************************************************************
# Copyright (c) 2021 Jittor. All Rights Reserved. 
# Maintainers: 
#     Guowei Yang <471184555@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import os, sys
import jittor as jt
import numpy as np
from jittor import nn
from jittor import dataset
mpi = jt.compile_extern.mpi


class Model(nn.Module):
    def __init__(self, input_size):
        self.linear1 = nn.Linear(input_size, 10)
        self.relu1 = nn.ReLU()
        self.linear2 = nn.Linear(10, 10)
    def execute(self, x):
        x = self.linear1(x)
        x = self.relu1(x)
        return self.linear2(x)

def fork_with_mpi(num_procs=4):
    import sys
    if jt.in_mpi:
        # you can mult other process output
        if jt.rank != 0:
            sys.stdout = open("/dev/null", "w")
        return
    else:
        print(sys.argv)
        cmd = " ".join(["mpirun", "-np", str(num_procs), sys.executable] + sys.argv)
        print("[RUN CMD]:", cmd)
        os.system(cmd)
        exit(0)

def main():
    mnist = dataset.MNIST()
    model = Model(mnist[0][0].size)
    sgd = jt.optim.SGD(model.parameters(), 1e-3)
    fork_with_mpi()

    for data, label in mnist:
        pred = model(data.reshape(data.shape[0], -1))
        # print(data.shape, label.shape, pred.shape)
        loss = nn.cross_entropy_loss(pred, label)
        sgd.step(loss)
        print(jt.rank, mnist.epoch_id, mnist.batch_id, loss)
        # break



# class TestMpiInPy(unittest.TestCase):
#     def test(self):
#         main()


if __name__ == "__main__":
    # unittest.main()
    main()


# ***************************************************************
# Copyright (c) 2021 Jittor. All Rights Reserved. 
# Maintainers: 
#     Guoye Yang <498731903@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import gc

def test(h, w, total_alloc_call, total_alloc_byte, total_free_call = 0, total_free_byte = 0):
        jt.clean()
        jt.gc()
        with jt.flag_scope(use_stat_allocator=1):
            a = jt.random([h,w])
            b = a+a
            c = a*b
            c.data
            del a,b,c
            gc.collect()
            x = (
                jt.flags.stat_allocator_total_alloc_call,
                jt.flags.stat_allocator_total_alloc_byte,
                jt.flags.stat_allocator_total_free_call,
                jt.flags.stat_allocator_total_free_byte
            )
            y = (total_alloc_call, total_alloc_byte, total_free_call, total_free_byte)
            assert x==y, (x, y)


class TestAllocator2(unittest.TestCase):
    def test_stat(self):
        #small_block
        test(10, 10, 1, 1048576) #800
        #small_block
        test(100, 100, 1, 1048576) #80000
        #large_block
        test(1000, 1000, 1, 20971520) #8000000
        #large_block2
        test(8000, 1000, 2, 67108864) #64000000

if __name__ == "__main__":
    unittest.main()



import os
import platform
import sys
import jittor_utils
from jittor_utils import LOG


def search_file(dirs, name):
    for d in dirs:
        fname = os.path.join(d, name)
        if os.path.isfile(fname):
            return fname
    LOG.f(f"file {name} not found in {dirs}")

if __name__ == "__main__":
    help_msg = f"Usage: {sys.executable} -m jittor_utils.config --include-flags|--link-flags|--cxx-flags|--cxx-example|--help"
    if len(sys.argv) <= 1:
        print(help_msg)
        sys.exit(1)

    s = ""
    # base should be something like python3.7m python3.8
    base = jittor_utils.get_py3_include_path().split()[0]
    base = "python3" + base.split("python3")[-1]
    for arg in sys.argv[1:]:
        if arg == "--include-flags":
            s += jittor_utils.get_py3_include_path()
            s += " -I"+os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "jittor", "src"))
            s += " "
        elif arg == "--libs-flags":
            libext = {
                'Linux': 'so',
                'Darwin': 'dylib',
                'Windows': 'DLL',
            }[platform.system()]
            ldflags = jittor_utils.run_cmd(jittor_utils.get_py3_config_path() + " --ldflags")
            libpaths = [l[2:] for l in ldflags.split(' ') if l.startswith("-L")]
            for libbase in libpaths:
                libpath = os.path.join(libbase, f"lib{base}.{libext}")
                if os.path.isfile(libpath):
                    s += f" -L{libbase} -l{base} -ldl "
                    break
            else:
                raise RuntimeError("Python dynamic library not found")
            if os.name == 'nt':
                s = s.replace('-ldl', '')
        elif arg == "--cxx-flags":
            s += " --std=c++17 -fPIC "
        elif arg == "--cxx-example":
            cc_src = '''
// please compile with: g++ a.cc $(python3 -m jittor_utils.config --include-flags --libs-flags --cxx-flags) -o a.out && ./a.out
#include <pyjt/pyjt_console.h>
#include <iostream>

using namespace std;

int main() {
    jittor::Console console;
    // run python code in console
    console.run("print('hello jt console', flush=True)");

    // set a python value: a = 1
    console.set<int>("a", 1);
    // get a python value
    cout << console.get<int>("a") << endl;

    // set a python string
    console.set<string>("b", "hello");
    cout << console.get<string>("b") << endl;

    // set a python array
    vector<int> x{1,2,3,4};
    console.set("x", x);
    auto x2 = console.get<std::vector<int>>("x");
    for (auto a : x2) cout << a << " "; cout << endl;

    // set and get a jittor array
    jittor::array<int, 2> arr2({2,3}, {6,5,4,3,2,1});
    arr2(0,0) = -1;
    console.set_array("arr2", arr2);
    console.run("print(arr2, flush=True); arr3 = arr2**2;");
    auto arr3 = console.get_array<int, 2>("arr3");
    cout << arr3.shape[0] << ' ' << arr3.shape[1] << endl;
    for (int i=0; i<arr3.shape[0]; i++) {
        for (int j=0; j<arr3.shape[1]; j++)
            cout << arr3(i,j) << ' ';
        cout << endl;
    }

    // run resnet18
    jittor::array<float, 4> input({2, 3, 224, 224});
    memset(input.data.get(), 0, input.nbyte());
    console.set_array("input", input);
    console.run(R"(
import jittor as jt
from jittor.models import resnet

model = resnet.resnet18()
pred = model(input)
    )");
    auto pred = console.get_array<float, 2>("pred");
    cout << "pred.shape " << pred.shape[0] << ' ' << pred.shape[1] << endl;

    return 0;
}
            '''
            print(cc_src)
        elif arg == "--help":
            print(help_msg)
            sys.exit(0)
        else:
            print(help_msg)
            sys.exit(1)
    print(s)



# ***************************************************************
# Copyright (c) 2021
#     Dun Liang <randonlang@gmail.com>. 
# All Rights Reserved.
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************

import jittor as jt
import unittest
import os
import numpy as np
import random

pass_this_test = False
msg = ""
mid = 0
if hasattr(os, "uname") and os.uname()[1] == "jittor-ce":
    mid = 1
try:
    # check can we run this test
    # test code
    jt.dirty_fix_pytorch_runtime_error()
    import torchvision.datasets as datasets
    import torchvision.transforms as transforms
    import torch

    traindir = ["/data1/cjld/imagenet/train/","/home/cjld/imagenet/train/"][mid]
    check_num_batch = 5
    assert os.path.isdir(traindir)

    normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                     std=[0.229, 0.224, 0.225])
    train_dataset = datasets.ImageFolder(
        traindir,
        transforms.Compose([
            transforms.RandomResizedCrop(224),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            normalize,
        ]))
except Exception as e:
    pass_this_test = True
    msg = str(e)

@unittest.skipIf(pass_this_test, f"can not run imagenet dataset test: {msg}")
class TestImageFolder(unittest.TestCase):
    def test_imagenet(self):
        train_loader = torch.utils.data.DataLoader(
            train_dataset, batch_size=256, shuffle=False)

        random.seed(0)
        tc_data = []
        for i, data in enumerate(train_loader):
            tc_data.append(data)
            print("get", data[0].shape)
            if i==check_num_batch: break

        from jittor.dataset.dataset import ImageFolder
        import jittor.transform as transform

        dataset = ImageFolder(traindir).set_attrs(batch_size=256, shuffle=False)

        dataset.set_attrs(transform = transform.Compose([
            transform.RandomCropAndResize(224),
            transform.RandomHorizontalFlip(),
            transform.ImageNormalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        ]))

        random.seed(0)

        for i, (images, labels) in enumerate(dataset):
            print("compare", i)
            assert np.allclose(images.numpy(), tc_data[i][0].numpy())
            assert np.allclose(labels.numpy(), tc_data[i][1].numpy())
            if i==check_num_batch: break

if __name__ == "__main__":
    unittest.main()
